import json
import subprocess
import sys
import tempfile


def execute_notebook(filename, suffix):
    with open(filename) as ff:
        nb_out = json.load(ff)

    source_code = [
        "".join(cell["source"]) + "\n\nplt.close()"
        for cell in nb_out["cells"]
        if cell["cell_type"] == "code"
    ]

    source_code = [
        "import matplotlib.pyplot as plt; plt.show = lambda: None"
    ] + source_code

    source_code = "\n".join(source_code)
    source_code = source_code.split("\n")
    source_code = [
        line
        for line in source_code
        if not line.startswith("%") and not line.startswith("!")
    ]
    source_code += [suffix]
    source_code = "\n".join(source_code)
    with tempfile.NamedTemporaryFile("w", delete=True) as f:
        with open(f.name, "w") as f:
            f.write(source_code)

        result = subprocess.check_output([sys.executable, f.name])
    return result.decode("utf-8")



from abc import ABC, abstractmethod
from typing import Callable, List, Tuple

from neurosym.python_dsl.python_ast_tools import fields_for_node

from .names import match_either_name_or_global
from .special_case_symbol_predicate import SpecialCaseSymbolPredicate


class Handler(ABC):
    """
    Corresponds to a given node in the SExpression AST. Keeps track of a list of defined
        production indices, in order of definition.
    """

    def __init__(self, mask, defined_production_idxs, config):
        assert isinstance(defined_production_idxs, list)
        self.mask = mask
        self.defined_production_idxs = defined_production_idxs
        self.config = config

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple["Handler", Callable[[], None]]:
        """
        When a child is entered, this method is called to determine the handler.

        Args:
            position: The position in the s-expression.
            symbol: The symbol of the child (index into a grammar's symbols list)
                Note: this can include variables, e.g., const-&x:0-Name,
                   but it can also include production symbols like Assign~S
                   or other leaves like const-i2~Const.

        Returns:
            The handler for the child.
            A function that can be called to undo the changes made by this function.
        """
        return (
            default_handler(
                position,
                symbol,
                self.mask,
                self.currently_defined_indices(),
                self.config,
            ),
            lambda: None,
        )

    def on_child_exit(
        self, position: int, symbol: int, child: "Handler"
    ) -> Callable[[], None]:
        """
        When a child is exited, this method is called to perform tasks related
            to the child.

        Returns:
            A function that can be called to undo the changes made by this function.
        """
        del position, symbol, child
        return lambda: None

    def currently_defined_indices(self) -> list[int]:
        """
        Returns the list of currently defined symbols, in order of definition.
        """
        assert isinstance(self.defined_production_idxs, list)
        return self.defined_production_idxs

    @abstractmethod
    def is_defining(self, position: int) -> bool:
        """
        Returns whether the context at the given position is defining.

        E.g., for an Assign node, the left-hand side is defining, and
            the right-hand side is not. This is important because for
            defining contexts, we do not need to use a previously
            defined variable.
        """

    def currently_defined_names(self):
        """
        Return the set of currently defined names. Note that this
            isn't the set of production symbols like const-&x:0-Name,
            but rather a set of names like x.
        """
        names = set()
        for symbol in self.currently_defined_indices():
            sym = self.mask.id_to_name(symbol)
            mat = match_either_name_or_global(sym)
            if not mat:
                continue
            names.add(mat.group("name"))
        return names

    def target_child(
        self, position: int, symbol: int
    ) -> Tuple["Handler", Callable[[], None]]:
        """
        Return a handler collecting targets for the given child.
        """
        # pylint: disable=cyclic-import
        from .target_handler import create_target_handler

        return (
            create_target_handler(
                position,
                symbol,
                self.mask,
                self.currently_defined_indices(),
                self.config,
            ),
            lambda: None,
        )

    def _matches(self, names, symbol_id, idx_to_name, special_case_predicates):
        """
        Whether or not the symbol matches the names.
        """

        for pred in special_case_predicates:
            if pred.applies(symbol_id):
                return pred.compute(symbol_id, names)
        if idx_to_name[symbol_id] is None:
            return True
        return idx_to_name[symbol_id] in names

    def compute_mask(
        self,
        position: int,
        symbols: List[int],
        idx_to_name: List[str],
        special_case_predicates: List[SpecialCaseSymbolPredicate],
    ):
        """
        Compute the mask for the given names. If the context is defining,
            then all symbols are valid. Otherwise, only the symbols that
            match the names are valid.
        """
        assert isinstance(special_case_predicates, list) and all(
            isinstance(x, SpecialCaseSymbolPredicate) for x in special_case_predicates
        )
        if self.is_defining(position):
            return [True] * len(symbols)
        names = set(self.currently_defined_names())
        return [
            self._matches(names, symbol, idx_to_name, special_case_predicates)
            for symbol in symbols
        ]


class ConstructHandler(Handler):
    """
    Handler for a single construct.
    """

    # must be overridden in subclasses, represents the name of the construct
    name: str = None

    def __init__(self, mask, defined_production_idxs, config):
        super().__init__(mask, defined_production_idxs, config)
        assert isinstance(self.name, str)
        self.child_fields = {
            field: i for i, field in enumerate(fields_for_node(self.name))
        }


class DefaultHandler(Handler):
    def is_defining(self, position: int) -> bool:
        return False


def default_handler(
    position: int, symbol: int, mask, defined_production_idxs, config
) -> Handler:
    # pylint: disable=cyclic-import
    from .defining_statement_handler import defining_statement_handlers

    assert isinstance(defined_production_idxs, list)

    symbol = mask.id_to_name(symbol)
    pulled = config.pull_handler(position, symbol, mask, defined_production_idxs)
    if pulled is not None:
        return pulled

    return defining_statement_handlers().get(symbol, DefaultHandler)(
        mask, defined_production_idxs, config
    )


class HandlerPuller(ABC):
    @abstractmethod
    def pull_handler(
        self,
        position: int,
        symbol: int,
        mask,
        defined_production_idxs,
        config,
        handler_fn: Callable,
    ) -> Handler:
        pass



from typing import List, Set

from s_expression_parser import Pair, ParserConfig, Renderer, nil, parse

from neurosym.programs.s_expression import SExpression


def to_pair(s_exp: SExpression, *, for_stitch: bool) -> Pair:
    """
    Convert an SExpression to a Pair.

    If we are exporting for stitch and the SExpression is a leaf,
        it will be converted to a string with the prefix "leaf-".
        This is because stitch does not distinguish `(f)` from `f`.

    Args:
        s_exp: The SExpression to convert.
        for_stitch: Whether the Pair is being converted for use with stitch.
    Returns:
        The Pair representing the SExpression.
    """
    if hasattr(s_exp, "__to_pair__"):
        return s_exp.__to_pair__(for_stitch=for_stitch)
    if isinstance(s_exp, str):
        return s_exp
    assert isinstance(s_exp, SExpression), f"Expected SExpression, got {s_exp}"
    if for_stitch and not s_exp.children:
        if s_exp.symbol.startswith("$"):
            return s_exp.symbol
        return "leaf-" + s_exp.symbol
    elements = [s_exp.symbol] + [
        to_pair(x, for_stitch=for_stitch) for x in s_exp.children
    ]
    result = nil
    for element in reversed(elements):
        result = Pair(element, result)
    return result


def from_pair(
    pair: Pair, should_not_be_leaf: Set[str], for_stitch: bool
) -> SExpression:
    """
    Convert a Pair to an SExpression.

    Args:
        pair: The Pair to convert.
        should_not_be_leaf: A set of symbols that should not be converted to leaves.
            Instead, they should be converted to SExpressions with no children.
        for_stitch: Whether the Pair is being converted for use with stitch.

    Returns:
        The SExpression representing the Pair.
    """
    if isinstance(pair, str):
        if for_stitch:
            if pair.startswith("leaf-"):
                return SExpression(pair[5:], ())
            if pair.startswith("$"):
                return SExpression(pair, ())
        if pair in should_not_be_leaf:
            return SExpression(pair, ())
        return pair
    assert isinstance(pair, Pair), f"Expected pair, got {pair}"
    elements = []
    while pair is not nil:
        car = pair.car
        if not elements:
            assert isinstance(car, str), f"Expected string, got {pair.car}"
        else:
            car = from_pair(car, should_not_be_leaf, for_stitch=for_stitch)
        elements.append(car)
        pair = pair.cdr
    head, *tail = elements
    return SExpression(head, tuple(tail))


def render_s_expression(s_exp: SExpression, for_stitch: bool = False) -> str:
    """
    Render an SExpression as a string.

    Args:
        s_exp: The SExpression to render.
        for_stitch: Whether the Pair is being converted for use with stitch.

    Returns:
        The string representing the SExpression.
    """
    return Renderer(columns=float("inf")).render(to_pair(s_exp, for_stitch=for_stitch))


def parse_s_expression(
    s: str, *, should_not_be_leaf: Set[str] = None, for_stitch: bool = False
) -> SExpression:
    """
    Parse a string into an SExpression.

    Args:
        s: The string to parse.
        should_not_be_leaf: A set of symbols that should not be converted to leaves.
            Instead, they should be converted to SExpressions with no children.
        for_stitch: Whether the Pair is being converted for use with stitch.

    Returns:
        The SExpression representing the string.
    """
    if should_not_be_leaf is None:
        should_not_be_leaf = set()
    pairs = parse(s, ParserConfig((), dots_are_cons=False))
    if len(pairs) != 1:
        raise ValueError(f"Expected one expression, got {len(pairs)}")
    return from_pair(
        pairs[0], should_not_be_leaf=should_not_be_leaf, for_stitch=for_stitch
    )


def symbols_for_program(s: SExpression) -> List[str]:
    return [s.symbol] + [sym for x in s.children for sym in symbols_for_program(x)]



from .dsls.sequential_differentiable_dsl import example_rnn_dsl
from .dsls.simple_differentiable_dsl import differentiable_arith_dsl
from .methods.near_example_trainer import NEARTrainer, NEARTrainerConfig
from .models.mlp import MLP, MLPConfig, mlp_factory
from .models.rnn import (
    RNNConfig,
    Seq2ClassRNN,
    Seq2SeqRNN,
    rnn_factory_seq2class,
    rnn_factory_seq2seq,
)
from .models.torch_program_module import TorchProgramModule
from .neural_dsl import NeuralDSL, PartialProgramNotFoundError, create_modules
from .search_graph import near_graph



import unittest

import neurosym as ns


class TestAllRules(unittest.TestCase):
    def test_basic_arith(self):
        t = ns.TypeDefiner()
        self.maxDiff = None
        self.assertEqual(
            {
                t("i"): [
                    ("+", [t("i"), t("i")]),
                    ("1", []),
                ]
            },
            ns.examples.basic_arith_dsl().all_rules(
                care_about_variables=False, valid_root_types=[t("i")]
            ),
        )



import re
from dataclasses import dataclass

from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR


@dataclass(frozen=True, eq=True, order=True)
class ExtraVar:
    """
    Used to represent an extra variable, not found in the tree distribution.

    Used in handling De Bruijn variables.
    """

    id: int

    @classmethod
    def from_name(cls, name):
        mat = canonicalized_python_name_leaf_regex.match(name)
        if mat:
            return cls(int(mat.group("var")))
        return None

    def leaf_name(self):
        # Using typ by default is a bit of a hack, since we should really be using use_type
        # however, this would require us to add a leaf for every version of use_type to the
        # tree distribution
        return canonicalized_python_name_as_leaf(self.id, "Name")


canonicalized_python_name_leaf_regex = re.compile(
    r"const-&(__(?P<var>\d+)):[0-9]+(" + re.escape(PYTHON_DSL_SEPARATOR) + r"[A-Za-z])?"
)


def canonicalized_python_name_as_leaf(name, typ):
    """
    Get the canonicalized python name as a leaf node. E.g., __0
    """
    result = f"const-&{canonicalized_python_name(name)}:0"
    result += PYTHON_DSL_SEPARATOR + typ
    return result


def canonicalized_python_name(name):
    return f"__{name}"



import unittest
from typing import Callable

import numpy as np
import scipy.stats

import neurosym as ns


class ProbabilityTester(unittest.TestCase):
    def assertBinomial(self, n, p, pdelta, k):
        plow = p - pdelta
        phigh = p + pdelta
        kmin = int(plow * n)
        kmax = int(phigh * n)

        prob_mass = scipy.stats.binom.cdf(kmax, n, p) - scipy.stats.binom.cdf(
            kmin - 1, n, p
        )
        self.assertTrue(kmin > 0, f"Expected kmin > 0, got {kmin}")
        self.assertTrue(kmax < n - 1)
        self.assertTrue(
            prob_mass > 0.99, f"Insufficient probability mass: {prob_mass:.3f}"
        )
        self.assertTrue(kmin <= k <= kmax, f"Expected {k} to be in [{kmin}, {kmax}]")

    def assertSameProbDist(self, dist, target):
        print(dist.distribution.tolist())
        np.testing.assert_almost_equal(dist.distribution, target)


class ChildrenInOrderMask(ns.PreorderMask):
    def __init__(self, tree_dist, dsl):
        del dsl
        super().__init__(tree_dist)
        self.proper_context = False
        self.seen_symbols = []

    def compute_mask(self, position, symbols):
        if self.proper_context:
            mask = []
            for symbol in symbols:
                symbol = self.tree_dist.symbols[symbol][0]
                mask.append(int(symbol) == len(self.seen_symbols) + 1)
            return mask
        return [True] * len(symbols)

    def on_entry(self, position, symbol) -> Callable[[], None]:
        symbol = self.tree_dist.symbols[symbol][0]
        undos = []
        if self.proper_context:
            self.seen_symbols.append(symbol)
            undos.append(self.seen_symbols.pop)
        if symbol == "+":
            previous_context = self.proper_context
            self.proper_context = True
            undos.append(lambda: setattr(self, "proper_context", previous_context))
        return ns.chain_undos(undos)

    def on_exit(self, position, symbol) -> Callable[[], None]:
        return lambda: None


class ChildrenInOrderAsserterMask(ns.PreorderMask):
    def __init__(self, tree_dist, dsl):
        del dsl
        super().__init__(tree_dist)
        self.proper_context = False
        self.seen_symbols = []

    def compute_mask(self, position, symbols):
        return [True] * len(symbols)

    def on_entry(self, position, symbol) -> Callable[[], None]:
        symbol = self.tree_dist.symbols[symbol][0]
        undos = []
        if self.proper_context:
            self.seen_symbols.append(symbol)
            undos.append(self.seen_symbols.pop)
            assert int(symbol) == len(
                self.seen_symbols
            ), f"Expected {len(self.seen_symbols)}, got {symbol}"
        if symbol == "+":
            previous_context = self.proper_context
            self.proper_context = True
            undos.append(lambda: setattr(self, "proper_context", previous_context))
        return ns.chain_undos(undos)

    def on_exit(self, position, symbol):
        symbol = self.tree_dist.symbols[symbol][0]
        if symbol == "+":
            assert self.proper_context, "Expected proper context"
            assert (
                len(self.seen_symbols) == 3
            ), f"Expected 3 children, but received {len(self.seen_symbols)}"
            previous_context = self.proper_context
            previous_symbols = self.seen_symbols
            self.proper_context = False
            self.seen_symbols = []

            def undo():
                self.proper_context = previous_context
                self.seen_symbols = previous_symbols

            return undo
        return lambda: None



"""
RNN example
"""
import torch
from torch import nn

from neurosym.dsl.dsl_factory import DSLFactory

from ..operations.basic import ite_torch
from ..operations.lists import fold_torch, map_torch


def example_rnn_dsl(L, O):
    dslf = DSLFactory(L=L, O=O, max_overall_depth=5)
    dslf.typedef("fL", "{f, $L}")

    dslf.concrete("add", "() -> ($fL, $fL) -> $fL", lambda: lambda x, y: x + y)
    dslf.concrete("mul", "() -> ($fL, $fL) -> $fL", lambda: lambda x, y: x * y)
    dslf.concrete(
        "fold", "((#a, #a) -> #a) -> [#a] -> #a", lambda f: lambda x: fold_torch(f, x)
    )
    dslf.concrete(
        "sum", "() -> $fL -> f", lambda: lambda x: torch.sum(x, dim=-1).unsqueeze(-1)
    )
    dslf.parameterized(
        "linear", "() -> $fL -> $fL", lambda lin: lin, dict(lin=lambda: nn.Linear(L, L))
    )
    dslf.parameterized(
        "output",
        "(([$fL]) -> [$fL]) -> [$fL] -> [{f, $O}]",
        lambda f, lin: lambda x: lin(f(x)),
        dict(lin=lambda: nn.Linear(L, O)),
    )
    dslf.concrete(
        "ite",
        "(#a -> f, #a -> #a, #a -> #a) -> #a -> #a",
        ite_torch,
    )
    dslf.concrete(
        "map", "(#a -> #b) -> [#a] -> [#b]", lambda f: lambda x: map_torch(f, x)
    )

    dslf.concrete(
        "compose", "(#a -> #b, #b -> #c) -> #a -> #c", lambda f, g: lambda x: g(f(x))
    )

    dslf.prune_to("[{f, $L}] -> [{f, $O}]")
    return dslf.finalize()



import ast
import os
import unittest

import parameterized


def files_to_examine():
    for root, _, files in os.walk("tests"):
        for file in files:
            if file.endswith(".py"):
                yield os.path.join(root, file)


class GatherImports(ast.NodeVisitor):
    def __init__(self):
        self.imports = set()

    def visit_Import(self, node):
        self.imports.add(node)

    def visit_ImportFrom(self, node):
        self.imports.add(node)


class OnlyDirectImportsTest(unittest.TestCase):
    @parameterized.parameterized.expand([(path,) for path in files_to_examine()])
    def test_only_direct_import(self, path):
        with open(path) as f:
            code = f.read()
        tree = ast.parse(code)
        gatherer = GatherImports()
        gatherer.visit(tree)
        imports = {ast.unparse(node) for node in gatherer.imports}
        imports = {imp for imp in imports if "neurosym" in imp}

        expected = {"from neurosym.examples import near", "import neurosym as ns"}

        self.assertEqual(imports | expected, expected)



import ast
import base64
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

from frozendict import frozendict
from increase_recursionlimit import increase_recursionlimit

from neurosym.programs.s_expression import SExpression

from .splice import Splice
from .symbol import PythonSymbol


class PythonAST(ABC):
    """
    Represents a Parsed AST.
    """

    @abstractmethod
    def to_ns_s_exp(self, config=frozendict()) -> SExpression:
        """
        Convert this PythonAST into a pair s-expression.
        """

    def to_python(self) -> str:
        """
        Convert this PythonAST into python code.
        """
        with increase_recursionlimit():
            code = self.to_python_ast()
            if isinstance(code, Splice):
                code = code.target
            return ast.unparse(code)

    @abstractmethod
    def to_python_ast(self) -> ast.AST:
        """
        Convert this PythonAST into a python AST.
        """

    @abstractmethod
    def map(self, fn):
        """
        Map the given function over this PythonAST. fn is run in post-order,
            i.e., run on all the children and then on the new object.
        """


@dataclass
class SequenceAST(PythonAST):
    head: str
    elements: List[PythonAST]

    def __post_init__(self):
        assert isinstance(self.head, str), self.head
        assert all(isinstance(x, PythonAST) for x in self.elements), self.elements

    def to_ns_s_exp(self, config=frozendict()):
        return SExpression(self.head, [x.to_ns_s_exp(config) for x in self.elements])

    def to_python_ast(self):
        result = []
        for x in self.elements:
            x = x.to_python_ast()
            if isinstance(x, Splice):
                result += x.target
            else:
                result += [x]
        return result

    def map(self, fn):
        return fn(SequenceAST(self.head, [x.map(fn) for x in self.elements]))


@dataclass
class NodeAST(PythonAST):
    typ: type
    children: List[PythonAST]

    def to_ns_s_exp(self, config=frozendict()):
        if not self.children and not config.get("no_leaves", False):
            return self.typ.__name__

        return SExpression(
            self.typ.__name__, [x.to_ns_s_exp(config) for x in self.children]
        )

    def to_python_ast(self):
        out = self.typ(*[x.to_python_ast() for x in self.children])
        out.lineno = 0
        return out

    def map(self, fn):
        return fn(NodeAST(self.typ, [x.map(fn) for x in self.children]))


@dataclass
class ListAST(PythonAST):
    children: List[PythonAST]

    def to_ns_s_exp(self, config=frozendict()):
        if not self.children:
            return SExpression("list", []) if config.get("no_leaves", False) else "nil"

        return SExpression("list", [x.to_ns_s_exp(config) for x in self.children])

    def to_python_ast(self):
        return [x.to_python_ast() for x in self.children]

    def map(self, fn):
        return fn(ListAST([x.map(fn) for x in self.children]))


@dataclass
class LeafAST(PythonAST):
    leaf: object

    def __post_init__(self):
        assert not isinstance(self.leaf, PythonAST)

    def to_ns_s_exp(self, config=frozendict()):
        leaf_as_string = self.render_leaf_as_string()
        if not config.get("no_leaves", False):
            return leaf_as_string
        return SExpression("const-" + leaf_as_string, [])

    def render_leaf_as_string(self):
        if (
            self.leaf is True
            or self.leaf is False
            or self.leaf is None
            or self.leaf is Ellipsis
        ):
            return str(self.leaf)
        if isinstance(self.leaf, PythonSymbol):
            return self.leaf.render()
        if isinstance(self.leaf, float):
            return f"f{self.leaf}"
        if isinstance(self.leaf, int):
            return f"i{self.leaf}"
        if isinstance(self.leaf, complex):
            return f"j{self.leaf}"
        if isinstance(self.leaf, str):
            # if all are renderable directly without whitespace, just use that
            if all(
                c in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_."
                for c in self.leaf
            ):
                return "s_" + self.leaf
            return "s-" + base64.b64encode(
                str([ord(x) for x in self.leaf]).encode("ascii")
            ).decode("utf-8")
        if isinstance(self.leaf, bytes):
            return "b" + base64.b64encode(self.leaf).decode("utf-8")
        raise RuntimeError(f"invalid leaf: {self.leaf}")

    def to_python_ast(self):
        if isinstance(self.leaf, PythonSymbol):
            return self.leaf.name
        return self.leaf

    def map(self, fn):
        return fn(LeafAST(self.leaf))


@dataclass
class SliceElementAST(PythonAST):
    content: PythonAST

    @classmethod
    def of(cls, x):
        if isinstance(x, SliceElementAST):
            return x
        return SliceElementAST(x)

    def to_ns_s_exp(self, config=frozendict()):
        # should not be necessary; since we have the assertion
        # but pylint is not smart enough to figure that out
        # pylint: disable=no-member
        content = self.content
        if isinstance(self.content, StarrableElementAST):
            # safe because it is not actually legal to have a starred element
            # in a slice
            content = content.content
        if isinstance(content, NodeAST):
            if content.typ is ast.Slice:
                return SExpression("_slice_slice", [content.to_ns_s_exp(config)])
            if content.typ is ast.Tuple:
                assert isinstance(content.children, list)
                assert len(content.children) == 2
                content_children = list(content.children)
                content_children[0] = ListAST(
                    [SliceElementAST.of(x) for x in content_children[0].children]
                )
                content = NodeAST(typ=ast.Tuple, children=content_children)

                return SExpression("_slice_tuple", [content.to_ns_s_exp(config)])
        return SExpression("_slice_content", [content.to_ns_s_exp(config)])

    def to_python_ast(self):
        return self.content.to_python_ast()

    def map(self, fn):
        return fn(SliceElementAST(self.content.map(fn)))


@dataclass
class StarrableElementAST(PythonAST):
    content: PythonAST

    def to_ns_s_exp(self, config=frozendict()):
        # pylint: disable=no-member
        if isinstance(self.content, NodeAST) and self.content.typ is ast.Starred:
            return SExpression("_starred_starred", [self.content.to_ns_s_exp(config)])
        return SExpression("_starred_content", [self.content.to_ns_s_exp(config)])

    def to_python_ast(self):
        return self.content.to_python_ast()

    def map(self, fn):
        return fn(StarrableElementAST(self.content.map(fn)))


@dataclass
class SpliceAST(PythonAST):
    content: PythonAST

    def to_ns_s_exp(self, config=frozendict()):
        return SExpression("/splice", [self.content.to_ns_s_exp(config)])

    def to_python_ast(self):
        return Splice(self.content.to_python_ast())

    def map(self, fn):
        return fn(SpliceAST(self.content.map(fn)))
